import os
from litellm import completion
from loguru import logger
from dotenv import load_dotenv
from art import *
import re
import markdown
from bs4 import BeautifulSoup
import html2text
from bs4 import BeautifulSoup, NavigableString

# import litellm
# litellm.set_verbose=True

class CommitCraftExtension(markdown.Extension):
    def extendMarkdown(self, md):
        md.preprocessors.register(CommitCraftPreprocessor(md), 'commit_craft', 175)

class CommitCraftPreprocessor(markdown.preprocessors.Preprocessor):
    def run(self, lines):
        new_lines = []
        in_commit_craft = False
        in_commit_msg = False
        for line in lines:
            if line.strip() == '```commit-craft':
                in_commit_craft = True
                new_lines.append('<div class="commit-craft">')
            elif line.strip() == '```commit-msg' and in_commit_craft:
                in_commit_msg = True
                new_lines.append('<div class="commit-msg">')
            elif line.strip() == '```' and in_commit_msg:
                in_commit_msg = False
                new_lines.append('</div>')
            elif line.strip() == '```' and in_commit_craft:
                in_commit_craft = False
                new_lines.append('</div>')
            else:
                new_lines.append(line)
        return new_lines

class CommitCraft:
    def __init__(self, model_name, stage_info_file, llm_output_file, html_output_file=None):
        self.model_name = model_name
        self.stage_info_file = stage_info_file
        self.llm_output_file = llm_output_file
        self.html_output_file = html_output_file
        
        tprint("CommitCraft")

    def html_to_markdown(self, element, indent=0):
        if isinstance(element, NavigableString):
            return element.strip()
        
        result = []
        if element.name == 'div' and 'commit-msg' in element.get('class', []):
            result.append('```commit-msg')
            for child in element.children:
                result.append(self.html_to_markdown(child, indent + 1))
            result.append('```')
        elif element.name in ['p', 'div']:
            for child in element.children:
                result.append(self.html_to_markdown(child, indent))
        elif element.name == 'ul':
            for li in element.find_all('li', recursive=False):
                result.append(f"{'  ' * indent}- {self.html_to_markdown(li, indent + 1)}")
        else:
            for child in element.children:
                result.append(self.html_to_markdown(child, indent))
        
        return '\n'.join(result)

    def extract_and_format_markdown_content(self, content):
        # Convert Markdown to HTML with custom extension
        html_content = markdown.markdown(content, extensions=[CommitCraftExtension()])
        
        # Save the HTML content to a file
        if(self.html_output_file):
            with open(self.html_output_file, "w", encoding="utf-8") as f:
                f.write(html_content)
        
        logger.success(f"HTML content saved to {self.html_output_file}")
        
        # Parse HTML
        soup = BeautifulSoup(html_content, 'html.parser')
        
        # Find the commit-craft block
        commit_craft_block = soup.find('div', class_='commit-craft')
        
        if commit_craft_block:
            # Convert HTML back to Markdown
            markdown_content = self.html_to_markdown(commit_craft_block)
            
            return markdown_content.strip()
        else:
            return None

    def generate_commit_messages(self):
        if self.model_name is None:
            logger.info("モデル名が指定されていません。CommitCraftをスキップします。")
            return

        with open(self.stage_info_file, "r", encoding="utf-8") as f:
            stage_info_content = f.read()
            
        logger.info(f"ステージ情報ファイル : '{self.stage_info_file}'")
        logger.info(f"モデル'{self.model_name}'を使用してLLMにステージ情報を送信しています...")
        response = completion(
            model=self.model_name, 
            messages=[{"role": "user", "content": stage_info_content}]
        )

        # レスポンスからコンテンツを抽出
        content = response.get('choices', [{}])[0].get('message', {}).get('content')

        # コンテンツからcommit-craftブロック内の内容を抽出し整形
        extracted_content = self.extract_and_format_markdown_content(content)
        
        if extracted_content:
            logger.info("抽出・整形されたcommit-craftブロックの内容:")
            logger.info(extracted_content)
            
            # 抽出・整形されたコンテンツを指定されたファイルに書き込む
            with open(self.llm_output_file, "w", encoding="utf-8") as f:
                f.write(extracted_content)
            
            logger.success(f"抽出・整形されたcommit-craftブロックの内容が{self.llm_output_file}に保存されました。")
        else:
            logger.info("commit-craftブロックは見つかりませんでした。")

# 使用例
if __name__ == "__main__":
    # 注意: この例では実際のLLM呼び出しは行いません
    commit_craft = CommitCraft("gpt-3.5-turbo", "stage_info.md", "llm_output.md", "output.html")
    
    # LLMレスポンスのシミュレーション
    mock_llm_response = {
        'choices': [{
            'message': {
                'content': '''ここにLLMの応答が入ります。例えば：

```commit-craft
# ✨ feat: ファイル変換機能を実装

## feature/file-converter

### app.py

```commit-msg
✨ feat: ファイル変換機能を実装

- Streamlitを用いて.py、.ipynb、.mdファイルを変換するアプリを作成しました。
- ユーザーは変換したいファイルと出力形式を選択できます。
- Jupytextライブラリを用いてファイル変換を行います。
- 変換されたファイルはダウンロードできます。
```
```

これが抽出すべき内容です。'''
            }
        }]
    }
    
    # モックレスポンスからコンテンツを抽出
    content = mock_llm_response.get('choices', [{}])[0].get('message', {}).get('content')
    
    # コンテンツからcommit-craftブロック内の内容を抽出し整形
    extracted_content = commit_craft.extract_and_format_markdown_content(content)
    
    if extracted_content:
        print("抽出・整形されたcommit-craftブロックの内容:")
        print(extracted_content)
        
        # 抽出・整形されたコンテンツを指定されたファイルに書き込む
        with open("llm_output.md", "w", encoding="utf-8") as f:
            f.write(extracted_content)
    else:
        print("commit-craftブロックは見つかりませんでした。")



# sourcesage\modules\IssueWize.py
import json
import os
import re
import time
from dotenv import load_dotenv
from litellm import completion
from loguru import logger

from art import *

try:
    from .GitCommander import run_command
except:
    from GitCommander import run_command
    
# .envファイルから環境変数を読み込む

# ロガーの設定
logger.add("issue_creator.log", rotation="1 MB", compression="zip", enqueue=True)

class IssueWize:
    def __init__(self, model="gemini/gemini-pro"):
        
        tprint("IssueWize")
        
        self.model = model

    def extract_json_from_response(self, response_text):
        """
        応答テキストからJSONのコードブロックを抽出する。
        コードブロックが見つからない場合は、応答テキスト全体をJSONとして扱う。

        Args:
            response_text (str): LiteLLMからの応答テキスト

        Returns:
            str: 抽出されたJSONの文字列
        """
        json_match = re.search(r'```json(.*?)```', response_text, re.DOTALL)
        if json_match:
            return json_match.group(1).strip()
        else:
            logger.warning("応答テキストにJSONのコードブロックが見つかりませんでした。そのままJSONとしてパースします。")
            return response_text

    def generate_issue_data(self, issue_summary, repo_overview):
        """
        issueの概要からタイトルと本文をJSONフォーマットで生成する。

        Args:
            issue_summary (str): issueの概要
            repo_overview (str): リポジトリの概要

        Returns:
            dict: 生成されたissueのデータ（タイトル、本文、ラベル）
        """
        issue_response = completion(
            model=self.model,
            messages=[{"role": "user", "content": f"""
以下のissueの概要からタイトルと本文とラベルをJSONフォーマットで生成してください:
下記のリポジトリの概要を参考にして生成して
タスクはなるべく分割して。
JSONのbodyはなるべく文章にして改行などは使用しないで句読点などで区切って。
ラベルは下記のラベルリストから参照して

{issue_summary}

## JSONフォーマットの例:
{{
  "title": "例：新機能の実装",
  "body": "例：新機能の詳細をマークダウン形式で記載します。",
  "labels": ["enhancement"]
}}

## ラベルリスト:

ラベルは以下から選択してください:
- bug: 何かが正常に動作していない 
- documentation: ドキュメントの改善や追加
- duplicate: 既に存在するIssueやPull Request 
- enhancement: 新機能やリクエスト
- good first issue: 初心者におすすめ
- help wanted: 特別な注意が必要
- invalid: 正しくないと思われるもの
- question: 詳細情報が求められている
- wontfix: 対応しないもの

## リポジトリの概要:
{repo_overview}

"""}]
        )
        issue_response_text = issue_response.get('choices', [{}])[0].get('message', {}).get('content')

        # 応答のテキストをファイルに保存
        with open("issue_response.txt", "w", encoding="utf-8") as f:
            f.write(issue_response_text)

        issue_json = self.extract_json_from_response(issue_response_text)

        try:
            # JSONをパースしてタイトルと本文を取得
            issue_data = json.loads(issue_json)
            issue_data['labels'] = issue_data.get('labels', [])
            return issue_data
        except json.JSONDecodeError as e:
            logger.error(f"JSONのデコードに失敗しました: {e}")
            logger.error("応答のテキストは'issue_response.txt'ファイルに保存されました。")
            raise

    def generate_detailed_markdown(self, issue_summary, issue_body):
        """
        issueの概要とBodyから詳細なマークダウンファイルを生成する。

        Args:
            issue_summary (str): issueの概要
            issue_body (str): LiteLLMから取得したissueのBody

        Returns:
            str: 生成された詳細なマークダウンファイルの内容
        """
        detailed_markdown_response = completion(
            model=self.model,
            messages=[{"role": "user", "content": f"""
以下のissueの概要とBodyから、詳細な要件定義やペルソナなどを記載したマークダウンファイルを生成してください。
マークダウンには章構成を多用して可読性を高めてください。

## Issueの概要:
{issue_summary}

## IssueのBody:
{issue_body}
"""}]
        )
        detailed_markdown = detailed_markdown_response.get('choices', [{}])[0].get('message', {}).get('content')
        return detailed_markdown

    def create_issue(self, issue_data, project, milestone):
        """
        ghコマンドを使用してissueを作成する。

        Args:
            issue_data (dict): issueのデータ（タイトル、本文、ラベル）
            project (str): プロジェクト名
            milestone (str): マイルストーン名
        """
        title = issue_data['title']
        body = issue_data['body']
        labels = issue_data['labels']

        # ghコマンドを使用してissueを作成
        command = [
            "gh", "issue", "create",
            "--title", title,
            "--body", body,
            "--project", project
        ]

        # コマンドにラベルを追加
        for label in labels:
            command.extend(["--label", label])

        # コマンドにマイルストーンを追加
        if milestone:
            command.extend(["--milestone", milestone])

        # run_command関数を使用してghコマンドを実行
        run_command(command)

    def create_optimized_issue(self, issue_summary, project, milestone, repo_overview_file, max_retries=3, retry_delay=5):
        """
        issueの概要から最適化されたissueを作成する。

        Args:
            issue_summary (str): issueの概要
            project (str): プロジェクト名
            milestone (str): マイルストーン名
            repo_overview_file (str): リポジトリ概要のマークダウンファイルパス
            max_retries (int): 最大リトライ回数（デフォルトは3）
            retry_delay (int): リトライ間の遅延時間（秒）（デフォルトは5秒）
        """
        retry_count = 0

        # リポジトリ概要のマークダウンファイルを読み込む
        with open(repo_overview_file, "r", encoding="utf-8") as f:
            repo_overview = f.read()

        while retry_count < max_retries:
            try:
                logger.info(f"Issueの作成を開始します。リトライ回数: {retry_count}")

                # LiteLLMを使用してissueの概要からタイトルと本文を生成
                issue_data = self.generate_issue_data(issue_summary, repo_overview)

                # 詳細なマークダウンファイルを生成
                detailed_markdown = self.generate_detailed_markdown(issue_summary, issue_data['body'])

                # 詳細なマークダウンファイルをissueのBodyとして設定
                issue_data['body'] = detailed_markdown

                # issueを作成
                self.create_issue(issue_data, project, milestone)

                logger.success("Issueの作成が完了しました。")
                break

            except Exception as e:
                logger.error(f"Issueの作成に失敗しました: {e}")
                retry_count += 1
                if retry_count < max_retries:
                    logger.warning(f"リトライします。リトライ回数: {retry_count}")
                    time.sleep(retry_delay)
                else:
                    logger.error("リトライ回数の上限に達しました。Issueの作成を中止します。")

if __name__ == '__main__':
    # 使用例
    issue_summary = r"""
Traceback (most recent call last):
  File "<frozen runpy>", line 198, in _run_module_as_main
  File "<frozen runpy>", line 88, in _run_code
  File "C:\Users\makim\miniconda3\Scripts\sourcesage.exe\__main__.py", line 7, in <module>
  File "C:\Users\makim\miniconda3\Lib\site-packages\sourcesage\cli.py", line 128, in main
    sourcesage.run()
  File "C:\Users\makim\miniconda3\Lib\site-packages\sourcesage\core.py", line 51, in run
    changelog_generator.integrate_changelogs()
  File "C:\Users\makim\miniconda3\Lib\site-packages\sourcesage\modules\ChangelogGenerator.py", line 150, in integrate_changelogs
    changelog_files = [file for file in os.listdir(self.output_dir) if file.startswith("CHANGELOG_")]
FileNotFoundError: [WinError 3] 指定されたパスが見つかりません。: './.SourceSageAssets/Changelog'

(base) C:\Prj\jupytext>
    """
    project_name = "TaskSphere"
    milestone = "Sprint01"
    repo_overview_file = r".SourceSageAssets\DOCUMIND\Repository_summary.md"
    # repo_overview_file = r"README.md"
    # model = "gemini/gemini-1.5-pro"
    model = "gemini/gemini-1.5-flash"
    issue_creator = IssueWize(model=model)
    issue_creator.create_optimized_issue(issue_summary, project_name, milestone, repo_overview_file, max_retries=5, retry_delay=10)



# sourcesage\cli.py
import argparse
from .core import SourceSage
from .modules.ReleaseDiffReportGenerator import GitDiffGenerator, MarkdownReportGenerator
import os
from loguru import logger
import sys
from art import *
import yaml

from .config.constants import Constants
from .modules.CommitCraft import CommitCraft
from .modules.DocuMind import DocuMind
from .modules.IssueWize import IssueWize

logger.configure(
    handlers=[
        {
            "sink": sys.stderr,
            "format": "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level:<10}</level> | <cyan>{name:<45}:{line:<5}</cyan> | <level>{message}</level>",
            "colorize": True,
        }
    ]
)

def add_arguments(parser):
    """SourceSage用の引数をパーサーに追加する"""
    
    # ==============================================
    # 基本設定
    #
    parser.add_argument('--ss-config', help='設定ファイルへのパス', default='sourcesage.yml')
    parser.add_argument('--ss-output', help='生成されたファイルの出力ディレクトリ', default='./')
    parser.add_argument('--repo', help='リポジトリへのパス', default='./')
    parser.add_argument('--owner', help='リポジトリのオーナー', default='Sunwood-ai-labs')  
    parser.add_argument('--repository', help='リポジトリの名前', default='SourceSage')  
    parser.add_argument('--ss-mode', nargs='+', help='実行するモード: Sage, GenerateReport, CommitCraft, DocuMind, IssueWize, またはall', default='all')
    # ==============================================
    # 変更履歴の設定
    #
    parser.add_argument('--changelog-start-tag', type=str, default=None, help='変更履歴の開始タグ')
    parser.add_argument('--changelog-end-tag', type=str, default=None, help='変更履歴の終了タグ')


    package_root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    default_ignore_file = os.path.join(package_root, 'sourcesage', 'config', '.SourceSageignore')
    default_language_map = os.path.join(package_root, 'sourcesage', 'config', 'language_map.json')
    
    # ==============================================
    # 無視ファイルと言語マップ設定
    #
    current_dir_ignore_file = os.path.join(os.getcwd(), '.SourceSageignore')
    if os.path.exists(current_dir_ignore_file):
        ignore_file = current_dir_ignore_file
        logger.info(f"カレントディレクトリの無視ファイルを使用: {ignore_file}")
    else:
        ignore_file = default_ignore_file
        logger.info(f"デフォルトの無視ファイルを使用: {ignore_file}")

    parser.add_argument('--ignore-file', help='無視ファイルへのパス', default=ignore_file)
    parser.add_argument('--language-map', help='言語マップファイルへのパス', default=default_language_map)
    
    # ==============================================
    # レポート生成用の引数を追加
    #
    parser.add_argument('--repo-path', type=str, default="", help='gitリポジトリへのパス')
    parser.add_argument('--git-fetch-tags', type=str, nargs='+', default=["git", "fetch", "--tags"], help='gitタグを取得するコマンド')
    parser.add_argument('--git-tag-sort', type=str, nargs='+', default=["git", "tag", "--sort=-creatordate"], help='gitタグをソートするコマンド')
    parser.add_argument('--git-diff-command', type=str, nargs='+', default=["git", "diff"], help='git diffを生成するコマンド')
    parser.add_argument('--report-title', type=str, default="Git Diff レポート", help='Markdownレポートのタイトル')
    parser.add_argument('--report-sections', type=str, nargs='+', default=["version_comparison", "diff_details"], help='レポートに含めるセクション')
    parser.add_argument('--ss-output-path', type=str, default=".SourceSageAssets/RELEASE_REPORT/", help='Markdownレポートの保存先フォルダ')
    parser.add_argument('--report-file-name', type=str, default="Report_{latest_tag}.md", help='Markdownレポートのファイル名。{latest_tag}は最新のタグに置換されます。')
    
    # ==============================================
    # IssueWize用の引数を追加
    #
    parser.add_argument('--issue-summary', type=str, default=None, help='Issueの概要')
    parser.add_argument('--project-name', type=str, default=None, help='IssueWizeのプロジェクト名')
    parser.add_argument('--milestone-name', type=str, default=None, help='IssueWizeのマイルストーン名')
    parser.add_argument('--repo-overview-file', type=str, default=None, help='リポジトリ概要のマークダウンファイルパス')
    parser.add_argument('--issuewize-model', type=str, default=None, help='IssueWizeで使用するモデル名')
    
    # ==============================================
    # CommitCraft用の引数を追加
    #
    parser.add_argument('--llm-output', type=str, default="llm_output.md", help='LLMレスポンスの出力ファイル')
    parser.add_argument('--ss-model-name', type=str, default=None, help='LLMのモデル名（デフォルト: None）')
    parser.add_argument('--stage-info-file', type=str, default=".SourceSageAssets\COMMIT_CRAFT/STAGE_INFO\STAGE_INFO_AND_PROMT_GAIAH_B.md", help='ステージファイルパス')
    parser.add_argument('--commit-craft-output', type=str, default=".SourceSageAssets/COMMIT_CRAFT/", help='CommitCraftの出力フォルダ')
    
    # ==============================================
    # DocuMind用の引数を追加
    #
    parser.add_argument('--docuMind-model', type=str, default=None, help='DocuMindで使用するLLMのモデル名')
    parser.add_argument('--docuMind-db', type=str, default=".SourceSageAssets/DOCUMIND/Repository_summary.md", help='DocuMindのデータベースファイルのパス')
    parser.add_argument('--docuMind-release-report', type=str, default=".SourceSageAssets/RELEASE_REPORT/Report_{latest_tag}.md", help='リリースレポートのパス。{latest_tag}は最新のタグに置換されます。')
    parser.add_argument('--docuMind-changelog', type=str, default=".SourceSageAssets/Changelog/CHANGELOG_release_{version}.md", help='変更履歴のパス。{version}はバージョンに置換されます。')
    parser.add_argument('--docuMind-output', type=str, default=".SourceSageAssets/DOCUMIND/RELEASE_NOTES.md", help='リリースノートの出力パス')
    parser.add_argument('--docuMind-prompt-output', type=str, default=".SourceSageAssets/DOCUMIND/_PROMPT.md", help='リリースノート作成のプロンプト')
    parser.add_argument('--repo-name', type=str, default="SourceSage", help='リポジトリの名前')
    parser.add_argument('--repo-version', type=str, default="v0.2.0", help='リポジトリのバージョン')
    
    # ==============================================
    # Configファイル
    #
    parser.add_argument('-f', '--yaml-file', help='設定をyamlファイルから読み込む', default=None)

def parse_arguments():
    """コマンドライン引数を解析する"""
    parser = argparse.ArgumentParser(description='SourceSage CLI')
    tprint(" SourceSage", font="rnd-xlarge")
    add_arguments(parser)
    return parser.parse_args()

def load_config_from_yaml(yaml_file=None):
    """YAMLファイルから設定を読み込み、argparse.Namespaceオブジェクトを返す"""
    default_config_file = 'sourcesage_config.yml'

    if yaml_file:
        config_file = yaml_file
    elif os.path.exists(default_config_file):
        config_file = default_config_file
    else:
        return None  # 設定ファイルが見つからない場合は None を返す

    logger.info(f"{config_file}を読み込みます...")
    with open(config_file, 'r') as f:
        yaml_args = yaml.safe_load(f)
        args_dict = {}  # args_dict を初期化
        for key, value in yaml_args.items():
            key = key.replace("-", "_")
            logger.debug(">> {: >30} : {: <20}".format(str(key), str(value)))
            args_dict[key] = value
        return argparse.Namespace(**args_dict)

def log_arguments(args):
    """引数の内容をログ出力する"""
    logger.info(f"---------------------------- パラメータ ----------------------------")
    args_dict = vars(args)
    for key, value in args_dict.items():
        key = key.replace("-", "_")
        logger.debug(">> {: >30} : {: <20}".format(str(key), str(value)))

def run(args=None):
      
    # -----------------------------------------------
    # SourceSageの実行
    if 'all' in args.ss_mode or 'Sage' in args.ss_mode:
        logger.info("SourceSageを起動します...")
        sourcesage = SourceSage(args.ss_config, args.ss_output, args.repo, args.owner, args.repository, args.ignore_file, args.language_map,
                                args.changelog_start_tag, args.changelog_end_tag)
        sourcesage.run()

    # -----------------------------------------------  
    # IssueWizeを使用してIssueを作成
    #
    if 'all' in args.ss_mode or 'IssueWize' in args.ss_mode:
        issuewize = IssueWize(model=args.issuewize_model)
        if args.issue_summary and args.project_name and args.repo_overview_file:
            logger.info("IssueWizeを使用してIssueを作成します...")
            issuewize.create_optimized_issue(args.issue_summary, args.project_name, args.milestone_name, args.repo_overview_file)
        else:
            logger.warning("IssueWizeの実行に必要なパラメータが指定されていません。--issue-summary, --project-name, --repo-overview-fileを指定してください。")
        

    # -----------------------------------------------
    # レポートの生成
    #
    if 'all' in args.ss_mode or 'GenerateReport' in args.ss_mode:
        logger.info("git diff レポートの生成を開始します...")
        git_diff_generator = GitDiffGenerator(args.repo_path, args.git_fetch_tags, args.git_tag_sort, args.git_diff_command)
        diff, latest_tag, previous_tag = git_diff_generator.get_git_diff()

        if(diff != None):
            report_file_name = args.report_file_name.format(latest_tag=latest_tag)
            os.makedirs(args.ss_output_path, exist_ok=True)
            output_path = os.path.join(args.ss_output_path, report_file_name)

            markdown_report_generator = MarkdownReportGenerator(diff, latest_tag, previous_tag, args.report_title, args.report_sections, output_path)
            markdown_report_generator.generate_markdown_report()

    # -----------------------------------------------
    # CommitCraftを使用してLLMにステージ情報を送信し、コミットメッセージを生成
    #
    if 'all' in args.ss_mode or 'CommitCraft' in args.ss_mode:
        stage_info_file = args.stage_info_file
        llm_output_file = os.path.join(args.commit_craft_output, args.llm_output)
        os.makedirs(args.commit_craft_output, exist_ok=True)
        commit_craft = CommitCraft(args.ss_model_name, stage_info_file, llm_output_file)
        commit_craft.generate_commit_messages()
    
    # -----------------------------------------------
    # DocuMindを使用してリリースノートを生成
    #
    if 'all' in args.ss_mode or 'DocuMind' in args.ss_mode:
        docuMind = DocuMind(args.docuMind_model, args.docuMind_db, args.docuMind_release_report, args.docuMind_changelog, args.repo_name, args.repo_version, args.docuMind_prompt_output)
        release_notes = docuMind.generate_release_notes()
        docuMind.save_release_notes(args.docuMind_output, release_notes)

    logger.success("プロセスが完了しました。")
    tprint("!! successfully !!", font="rnd-medium")

def main():
    _args = parse_arguments() 
    args = load_config_from_yaml(_args.yaml_file) or _args 
    log_arguments(args) 
    run(args)

if __name__ == '__main__':
    main()
